import os

from ._internal import DEFAULT, defaultarguments, doreahconfig

_config = {}

# set configuration
# defaultextension	files with this extension will be regarded as valid files. can be overwritten per request.
# comments			whether files may include commenting (indicated by #)
# multitab			whether fields can be separated by multiple tabs (this makes empty fields impossible except when trailing)
def config(defaultextension=".tsv",comments=True,multitab=True):
	"""Configures default values for this module.

	These defaults define behaviour of function calls when respective arguments are omitted. Any call of this function will overload the configuration in the .doreah file of the project. This function must be called with all configurations, as any omitted argument will reset to default, even if it has been changed with a previous function call."""
	global _config
	_config["defaultextension"] = defaultextension
	_config["comments"] = comments
	_config["multitab"] = multitab


# initial config on import, set everything to default
config()


@defaultarguments(_config,comments="comments",multitab="multitab")
def parse(filename,*args,comments=DEFAULT,multitab=DEFAULT):
	"""Parses a tsv-like file.

	:param string filename: File to be parsed
	:param string args: Data types to be read
	:param boolean comments: Whether the file may contain comments
	:param boolean multitab: Whether values can be separated by an arbitrary amount of tabs
	:return: List of entries, one entry per line, each entry a list of fields"""

	if not os.path.exists(filename):
		filename = filename + _config["defaultextension"]

	f = open(filename)

	result = []
	for l in [l for l in f if (not l.startswith("#")) and (not l.strip()=="")]:
		l = l.replace("\n","")

		# if the file allows comments, we need to replace the escape sequence and properly stop parsing for inline comments
		if comments:
			l = l.split("#")[0]
			l = l.replace(r"\num","#")
			l = l.replace(r"\hashtag","#")

		# we either allow multiple tabs, or we don't (in which case empty fields are possible)
		if multitab:
			data = list(filter(None,l.split("\t")))
		else:
			data = list(l.split("\t"))

		entry = [] * len(args)
		for i in range(len(args)):
			if args[i] in ["list","ls","array"]:
				try:
					entry.append(data[i].split("␟"))
				except:
					entry.append([])
			elif args[i] in ["string","str","text"]:
				try:
					entry.append(data[i])
				except:
					entry.append("")
			elif args[i] in ["int","integer","num","number"]:
				try:
					entry.append(int(data[i]))
				except:
					entry.append(0)
			elif args[i] in ["bool","boolean"]:
				try:
					entry.append((data[i].lower() in ["true","yes","1","y"]))
				except:
					entry.append(False)
			else:
				raise TypeError()

		result.append(entry)

	f.close()
	return result

@defaultarguments(_config,extension="defaultextension")
def parse_all(path,*args,extension=DEFAULT,**kwargs):
	"""Parses all valid files in a specific directory

	:param string path: Directory to search for files
	:param string args: Data types to be read
	:param string extension: Only parse files ending with this string
	:param kwargs: Parse settings to be applied, see :func:`parse`"""

	result = []
	for f in os.listdir(path + "/"):
		if (f.endswith(extension)): # use "" if all files are valid
			result += parse(path + "/" + f,*args,**kwargs)

	return result



def create(filename):
	"""Creates a file if it doesn't already exist

	:param string filename: Full filename"""

	if not os.path.exists(filename):
		open(filename,"w").close()

@defaultarguments(_config,comments="comments")
def add_entry(filename,entry,comments=DEFAULT):
	"""Adds an entry to a tsv-like file

	:param string filename: Full filename
	:param iterable entry: List or tuple of fields
	:param boolean comments: Whether the file can contain comments"""

	create(filename)
	# remove all tabs and create tab-separated string
	line = "\t".join([str(field).replace("\t"," ") for field in entry])

	# replace comment symbol
	if comments: line = line.replace("#",r"\num")

	with open(filename,"a") as f:
		f.write(line + "\n")

@defaultarguments(_config,comments="comments")
def add_entries(filename,entries,comments=DEFAULT):
	"""Adds several entries to a tsv-like file

	:param string filename: Full filename
	:param iterable entries: List or tuple of entries, each a list of tuple of fields
	:param boolean comments: Whether the file can contain comments"""

	create(filename)

	with open(filename,"a") as f:
		for entry in entries:
			line = "\t".join([str(field).replace("\t"," ") for field in entry])
			if comments: line = line.replace("#",r"\num")
			f.write(line + "\n")



# now check local configuration file
_config.update(doreahconfig("tsv"))



import time

from ._internal import defaultarguments, doreahconfig

_config = {}

_lastcalls = {}

# set configuration
# si		0 means seconds, 1 ms, 2 μs, 3 ns etc
def config(si=0):
	"""Configures default values for this module.

	These defaults define behaviour of function calls when respective arguments are omitted. Any call of this function will overload the configuration in the .doreah file of the project. This function must be called with all configurations, as any omitted argument will reset to default, even if it has been changed with a previous function call."""
	global _config
	_config["si"] = si


# initial config on import, set everything to default
config()


# Take clock. Returns time passed since last call of this function. if called with an identifier, will only
# consider calls with that identifier. No identifier means any call is valid.
# identifiers	arbitrary strings to remember different timers. guaranteed to set all timers to exactly the same time for
#				all identifiers in one call. will return tuple of all identifiers, singular value if only one identifier
def clock(*identifiers):
	"""Takes the time. Returns elapsed time since last call with that identifier for each identifier,
	or time since last call with any identifier if no identifier is given.

	:param string identifiers: Unique identifiers for timing to be taken
	:return: Tuple of elapsed times for each identifier, or single value if only one identifier or no identifier."""

	global _lastcalls

	if len(identifiers) == 0: identifiers = (None,)

	now = time.time()
	# get last calls
	stamps = (_lastcalls.get(i) for i in identifiers)
	results = tuple(None if lc is None else (now - lc) * (1000**_config["si"]) for lc in stamps)
	if len(results) == 1: results = results[0]

	# set new stamps
	for i in identifiers:
		_lastcalls[i] = now
	_lastcalls[None] = now			# always save last overall call so we can directly access it

	return results



def clockp(name,*identifiers):
	"""Prints out elapsed time since last call according to the rules of :meth:`clock`

	:param string name: Name to be printed out
	:param string identifiers: Unique identifiers for timing to be taken"""
	time = clock(*identifiers)
	print(name + ": " + str(time))






# now check local configuration file
_config.update(doreahconfig("timing"))



import os
from functools import wraps

# This is purely for documentation
class Default:
	def __repr__(self):
		return u"<Module default>"
DEFAULT = Default()

## decorator to set default arguments that are only evaluated at runtime
def defaultarguments(defaultdict,**defaultargs):
	def decorator(func):	#actual decorator function
		@wraps(func)
		def newfunc(*args,**kwargs):
			realargs = {}

			# evaluate default values at runtime
			for k in defaultargs:
				realargs[k] = defaultdict[defaultargs[k]]

			# overwrite given arguments of function
			realargs.update(kwargs)

			# execute function with correct arguments
			return func(*args,**realargs)

		return newfunc

	return decorator


## opens file, creates it if necessary (including all folders)
def gopen(filepath,mode):
	directory = os.path.dirname(filepath)
	os.makedirs(directory, exist_ok=True)

	return open(filepath,mode)



# reads module configuration from file
def doreahconfig(module):
	from .settings import get_settings
	s = get_settings(files=[".doreah"],prefix=module + ".",cut_prefix=True)
	return s



### change decorator, use default object instead (sentinel) for better documentation



import pickle
import os
import math
import zlib
import random

from ._internal import DEFAULT, defaultarguments, gopen, doreahconfig

_config = {}

# set configuration
# folder	folder to store log files
def config(folder="storage"):
	"""Configures default values for this module.

	These defaults define behaviour of function calls when respective arguments are omitted. Any call of this function will overload the configuration in the .doreah file of the project. This function must be called with all configurations, as any omitted argument will reset to default, even if it has been changed with a previous function call."""
	global _config
	_config["folder"] = folder


# initial config on import, set everything to default
config()


@defaultarguments(_config,folder="folder")
def save(data,name,folder=DEFAULT):
	"""Saves the supplied data structure to disk.

	:param data: Object to be serialized and saved to disk
	:param string name: File name to be used
	:param string folder: Custom folder to save the file in"""

	filename = os.path.join(folder,name)

	try:
		with gopen(filename,"wb") as fl:
			stream = pickle.dumps(data)
			fl.write(stream)
	except:
		pass

@defaultarguments(_config,folder="folder")
def load(name,folder=DEFAULT):
	"""Loads a data structure from disk.

	:param string name: File name to be read
	:param string folder: Custom folder where the file is stored
	:return: Data object"""

	filename = os.path.join(folder,name)

	try:
		with gopen(filename,"rb") as fl:
			ob = pickle.loads(fl.read())
	except: ob = None

	return ob

@defaultarguments(_config,folder="folder")
def delete(name,folder=DEFAULT):
	"""Deletes a serialized data structure from disk.

	:param string name: File name
	:param string folder: Custom folder where the file is stored"""

	filename = os.path.join(folder,name)
	try:
		os.remove(filename)
	except:
		pass

@defaultarguments(_config,folder="folder")
def size(name,folder=DEFAULT):
	"""Returns the size of a serialized object.

	:param string name: File name
	:param string folder: Custom folder where the file is stored"""

	filename = os.path.join(folder,name)
	return os.path.getsize(filename)


# need a deterministic hash function
def _hash(val):
	if isinstance(val,int):
		return val
	elif isinstance(val,str):
		return zlib.adler32(bytes(val,encoding='UTF-8'))
	elif isinstance(val,dict):
		return _hash(tuple((k,val[k]) for k in sorted(val)))
	elif isinstance(val,tuple) or isinstance(val,list):
		x = 10
		i = 1
		for element in val:
			x = ((x * _hash(element)) ** i)
			i += 1
		return x

	try:
		return zlib.adler32(val)
	except: pass
	try:
		return zlib.adler32(bytes(val))
	except: pass



class DiskDict:
	"""Dictionary-like object that stores its authorative information completely on disk.

	:param integer maxmemory: Soft memory limit (in bytes), not meant for precision
	:param integer maxstorage: Soft disk space limit (in bytes), not meant for precision
	:param string name: Directory name used for storage
	:param string folder: Parent directory"""

	@defaultarguments(_config,folder="folder")
	def __init__(self,maxmemory=math.inf,maxstorage=1024*1024*1024*4,name="diskdict",folder=DEFAULT):

		self.maxmemory = maxmemory
		self.maxstorage = maxstorage

		self.name = name
		self.folder = os.path.join(folder,name)


		self.cache = {}
		self.cache_unhashable = {}


	def __contains__(self,key):
		try:
			self.get(key)
			return True
		except:
			return False
	def __getitem__(self,key):
		return self.get(key)
	def __setitem__(self,key,value):
		return self.add(key,value)
	def __delitem__(self,key):
		pass

	def __str__(self):
		return "{" + ",".join([str(key) + ":" + str(self.cache[key]) for key in self.cache] + ["etc."]) + "}"

	def get(self,key):
		"""Get the value of a key.

		:param key: Key to be retrieved
		:return: Value of the requested key
		:raises KeyError: No valid entry for the key found."""

		hashvalue = str(_hash(key))
		try:
			return self.cache[key]
		except: pass
		try:
			return self.cache_unhashable[hashvalue]
		except: pass

		if os.path.exists(os.path.join(self.folder,hashvalue)):
			possibilities = os.listdir(os.path.join(self.folder,hashvalue))
			for p in possibilities:
				result = load(os.path.join(self.name,hashvalue,p),folder=self.folder)
				if result[0] == key:
					val = result[1]
					try:
						self.cache[key] = val
					except TypeError:
						self.cache_unhashable[hashvalue] = val
					return val

		raise KeyError()



	def add(self,key,value):
		"""Add an entry.

		:param key: Key to be added
		:param value: Value of this entry"""

		hashvalue = str(_hash(key))

		# cache
		try:
			self.cache[key] = value
		except TypeError:
			self.cache_unhashable[hashvalue] = value

		name = str(random.uniform(1000000000,9999999999)).replace(".","")
		save((key,value),os.path.join(self.name,hashvalue,name),folder=self.folder)




# now check local configuration file
_config.update(doreahconfig("persistence"))



import os
import shutil

from ._internal import DEFAULT, defaultarguments, doreahconfig


_config = {}

# set configuration
# defaultextension	unused
# files				list of all files that will be used for configuration. high indicies overwrite low indicies
# comment			symbols that indicate comments. careful!
# category			symbols that indicate start and end of category name. only works at start of line
# onlytext			interpret everything as a string. if False, strings can be put into quotes to avoid confusion
def config(defaultextension=".ini",files=["settings.ini","settings.conf","configuration.ini","configuration.conf"],
			comment=["#","//"],category=("[","]"),onlytext=False):
	"""Configures default values for this module.

	These defaults define behaviour of function calls when respective arguments are omitted. Any call of this function will overload the configuration in the .doreah file of the project. This function must be called with all configurations, as any omitted argument will reset to default, even if it has been changed with a previous function call."""
	global _config
	_config["defaultextension"] = defaultextension
	_config["files"] = files
	_config["comment"] = comment
	_config["category"] = category
	_config["onlytext"] = onlytext

# initial config on import, set everything to default
config()





def _interpret(text):
	if _config["onlytext"]: return text

	if text.lower() in ["true","yes"]: return True
	if text.lower() in ["false","no"]: return False
	if text.lower() in ["none","nan","n/a",""]: return None
	if text.startswith("[") and text.endswith("]"):
		list = []
		buffer = ""
		string = None
		stringended = False
		for c in text[1:-1]:
			if stringended and c != ",": pass	#after a string is done, skip to the next delimiter
			elif c == '"' and string is None:
				string = '"'		# start new string
				buffer = '"'
			elif c == "'" and string is None:
				string = "'"		# start new string
				buffer = "'"
			elif c == '"' and string is '"':
				string = None		# terminate string
				stringended = True
				buffer += '"'
			elif c == "'" and string is "'":
				string = None		# terminate string
				stringended = True
				buffer += "'"
			elif c == "," and string is None:
				list.append(buffer)
				buffer = ""
				stringended = False
			else: buffer += c

		list.append(buffer.strip())
		return [_interpret(entry) for entry in list]

	if text.startswith("'") and text.endswith("'"): return text[1:-1]
	if text.startswith('"') and text.endswith('"'): return text[1:-1]
	try:
		return int(text)
	except:
		pass
	try:
		return float(text)
	except:
		pass
	return text


# get settings
# keys			list of requested keys. if present, will return list of according values, if not, will return dict of key-value pairs
# files			which files to parse. later files (higher indicies) will overload earlier ones
# prefix		only request keys with a certain prefix. key filter will still be applied if present
# cut_prefix	return keys without the prefix
# category		return only keys of specific category
# raw			do not interpret data type, only return strings
@defaultarguments(_config,files="files")
def get_settings(*keys,files=DEFAULT,prefix="",cut_prefix=False,category=None,raw=False):
	"""Get the active settings value for all supplied keys. Get a dict of all settings (or only filtered) if no keys are supplied.

	:param string keys: Setting keys to be read
	:param list files: List of settings files from least to most authorative
	:param string prefix: Only return settings keys beginning with this string
	:param boolean cut_prefix: Drop specified prefix from key names
	:param string category: Only return settings of this category
	:param boolean raw: Do not parse data types, return all values as strings
	:return: Tuple of values for all specified keys. If no keys are specified, dict of all settings that satisfy the constraints."""

	allsettings = {}

	for f in files:
		if not os.path.exists(f): continue

		# if category is specified, ignore all entries by default and switch when we encounter the right heading
		ignore = False if category is None else True

		with open(f) as file:
			lines = file.readlines()


		for l in lines:
			# clean up line
			l = l.replace("\n","")
			for symbol in _config["comment"]:
				l = l.split(symbol)[0]
			l = l.strip()

			# check if valid settings entry
			if l == "": continue
			if l.startswith(_config["category"][0]):
				# ignore category headers if we don't care
				if category is None: continue
				# otherwise, find out if this is the category we want
				else:
					if l.endswith(_config["category"][1]):
						cat = l[len(_config["category"][0]):-len(_config["category"][1])]
						ignore = not (cat == category) #if this is the right heading, set ignore to false
					continue

			if ignore: continue

			if "=" not in l: continue

			# read
			(key,val) = l.split("=")
			key = key.strip()
			val = val.strip() if raw else _interpret(val.strip())

			if key.startswith(prefix):
				# return keys without the common prefix
				if cut_prefix:
					allsettings[key[len(prefix):]] = val

				# return full keys
				else:
					allsettings[key] = val

	# no arguments means all settings
	if len(keys) == 0:
		return allsettings

	# specific keys requested
	else:
		if len(keys) == 1: return allsettings.get(keys[0])
		else: return [allsettings.get(k) for k in keys]


def update_settings(file,settings,create_new=False):
	"""Updates all settings in a file with settings from the supplied dictionary.

	:param string file: Settings file to write to
	:param dictionary settings: Settings data to write
	:param boolean create_new: Whether to write settings that are not already present in the file"""

	if not os.path.exists(file): open(file,"w").close()

	with open(file,"r") as origfile:
		lines = origfile.readlines()

	newlines = []

	for origline in lines:
		l = origline
		# clean up line
		l = l.replace("\n","")
		for symbol in _config["comment"]:
			l = l.split(symbol)[0]
		l = l.strip()

		# check if valid settings entry
		if l == "":
			newlines.append(origline)
			continue
		if l.startswith(_config["category"][0]):
			newlines.append(origline)
			continue
		if "=" not in l:
			newlines.append(origline)
			continue

		# read
		(key,val) = l.split("=")
		key = key.strip()
		val = val.strip()


		if key not in settings:
			newlines.append(origline)
			continue

		else:
			#print("Found key")
			newline = origline.split("=",1)
			#print({"linepart":newline[1],"keytoreplace":val,"new":settings[key]})
			newline[1] = newline[1].replace(val,str(settings[key]),1)
			newline = "=".join(newline)
			newlines.append(newline)

			del settings[key]

	if create_new:
		# settings that were not present in the file
		for key in settings:
			newlines.append(key + " = " + str(settings[key]) + "\n")

	with open(file,"w") as newfile:
		newfile.write("".join(newlines))





def update(source="default_settings.ini",target="settings.ini"):
	"""Updates a user settings file to a newer format from a default file without overwriting user's settings.

	:param string source: Default settings file that provides the format
	:param string target: Local settings file that provides higher authority on settings values"""

	if not os.path.exists(target):
		shutil.copyfile(source,target)
	else:
		usersettings = get_settings(files=[target],raw=True)
		shutil.copyfile(source,target)
		update_settings(target,usersettings)






# now check local configuration file
_config.update(doreahconfig("settings"))



from lxml import etree
from copy import deepcopy
import re
import os

from ._internal import DEFAULT, defaultarguments, doreahconfig

_config = {}


# set configuration
def config(interpret=str):
	"""Configures default values for this module.

	These defaults define behaviour of function calls when respective arguments are omitted. Any call of this function will overload the configuration in the .doreah file of the project. This function must be called with all configurations, as any omitted argument will reset to default, even if it has been changed with a previous function call."""
	global _config
	_config["interpret"] = interpret

# initial config on import, set everything to default
config()





@defaultarguments(_config,interpret="interpret")
def file(path,d={},interpret=DEFAULT,noroot=False):
	"""Parses a pyhp source file and returns the generated html code.

	:param string path: Location of the pyhp source file
	:param dict d: Variables dictionary
	:param function interpret: Function that determines string representations of variables
	:return: HTML source
	"""

	with open(path,"r") as f:
		content = f.read()

	directory = os.path.dirname(os.path.abspath(path))

	return parse(content,d,interpret=interpret,directory=directory,noroot=noroot)


def _file(path,d,interpret=DEFAULT,noroot=False):

	with open(path,"r") as f:
		content = f.read()

	directory = os.path.dirname(os.path.abspath(path))

	return _parse(content,d,interpret=interpret,directory=directory,noroot=noroot)


@defaultarguments(_config,interpret="interpret")
def parse(src,d={},interpret=DEFAULT,directory=None,noroot=False):
	"""Parses pyhp source and returns the generated html code.

	:param string src: Source string
	:param dict d: Variables dictionary
	:param function interpret: Function that determines string representations of variables
	:return: HTML source
	"""

	doc = _parse(src,d,interpret=interpret,directory=directory,noroot=noroot)


	#raw = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
	raw = etree.tostring(doc,encoding="unicode",method="html").replace("&gt;",">").replace("&lt;","<")


	return raw
	# yes I know
	# it's a temporary thing okay

# to tree
def _parse(src,d,interpret=DEFAULT,directory=None,noroot=False):

	if noroot:
		# pseudo root
		src = "<pseudoroot>" + src + "</pseudoroot>"

	doc = etree.XML(src)
	doc = _parse_node(doc,d,interpret,directory=directory)[0]

	if noroot:
		return list(doc) #all child elements

	return doc


def _parse_node(node,d,interpret,directory=None):


	## fix
	#if node.tag == "html":
	#	node.attrib["xmlns"] = "http://www.w3.org/1999/xhtml"


	## parse pyhp nodes

	if node.tag == "pyhp":

		### CODE

		if len(node.attrib) == 0:
			print("Executing code!")
			code = node.text
			code = code.strip("\t").strip(" ")
			code = code.split("\n")
			if code[0] != "" or code[-1] != "":
				print("Malformed code block!")
			code = code[1:-1]
			roottabs = 0
			for char in code[0]:
				if char == "\t" or char == " ":
					roottabs += 1
				else:
					break

			code = [line[roottabs:] for line in code]
			code = "\n".join(code)

			exec(code,d)

			return [node.tail]

		#### SAVE

		if _attr(node,"save") is not None and _attr(node,"as") is not None:
			#print("d[" + _attr(node,"as") + "] = " + _attr(node,"save"))
			d[_attr(node,"as")] = eval(_attr(node,"save"),d)

			return [node.tail]



		#### INCLUDE

		elif _attr(node,"include") is not None:
			filename = _attr(node,"include")
			if directory is None:
				# relative to execution path
				filename = filename
			else:
				# relative to this file
				filename = os.path.join(directory,filename)

			#with open(filename,"r") as f:
			#	raw = f.read()

			subnodes = _file(filename,d,interpret=interpret,noroot=True)

			return subnodes + [node.tail]


		#### IF

		elif _attr(node,"if") is not None:
			if eval(_attr(node,"if"),d):
				nodestoreturn = [node.text]
				for sn in node:
					nodestoreturn += _parse_node(sn,d,interpret,directory=directory)
				nodestoreturn += [node.tail]
				return nodestoreturn
			else:
				return [node.tail]


		#### FOR IN

		elif _attr(node,"for") is not None and _attr(node,"in") is not None:
			nodestoreturn = []
			# for loop of the elements
			first = True
			try:
				elements = eval(_attr(node,"in"),d)
			except:
				# allow invalid expressions in for loops, just ignore them
				elements = []
			for element in elements:
				if not first and _attr(node,"separator") is not None:
					nodestoreturn += [_attr(node,"separator")]
				first = False

				# now go through the nodes each time
				nodestoreturn += [node.text]

				for sn in node:

					# parse with the normal dict and the current element of the loop
					sn = deepcopy(sn)
					nodestoreturn += _parse_node(sn,{**d, **{_attr(node,"for"):element}},interpret,directory=directory)



			nodestoreturn += [node.tail]

			return nodestoreturn


		#### ECHO

		elif _attr(node,"echo") is not None:
			#return [d.get(node.get("echo"))]
			return [interpret(eval(_attr(node,"echo"),d))] + [node.tail]

		return []


	## parse normal nodes

	else:

		## replace attributes (not necessary in pyhp nodes)

		for name,value in node.attrib.items():
			vars = re.findall("{.*?}",value)
			for v in vars:
				vname = v[1:-1]
				try:
					value = value.replace(v,interpret(eval(vname,d)))
				except:
					pass
					print("Error parsing:",v,"in attribute")

			node.attrib[name] = value



		subnodes = [n for n in node]
		newsubnodes = []
		for subnode in subnodes:
			node.remove(subnode)
			newsubnodes += _parse_node(subnode,d,interpret,directory=directory)

		prev = None
		for nsn in newsubnodes:


			if nsn is None: continue

			if isinstance(nsn,str):

				# string nodes need to be either the tail of their previous node, or if they're
				# in first position, the text of the parent node
				if prev is None:
					#if node.tag == "head": print(str(nsn) + " appended as text to " + str(node))
					try:
						node.text += nsn
					except:
						node.text = nsn
				else:
					#if node.tag == "head": print(str(nsn) + " appended as tail to " + str(prev))
					try:
						prev.tail += nsn
					except:
						prev.tail = nsn
			else:
				# once we encounter a real node, we can set it as prev for all following text nodes
				prev = nsn



		for nsn in newsubnodes:
			if isinstance(nsn,etree._Element):
				node.append(nsn)


		return [node]



def _attr(node,name):
	res = node.get(name)
	try:
		return res.replace(" gr "," > ").replace(" ls "," < ")
	except:
		return res



# now check local configuration file
_config.update(doreahconfig("pyhp"))





### run test server

if __name__ == "__main__":
	from bottle import get, run, static_file
	from doreah.pyhp import file
	import os

	@get("/<path:path>")
	def serve_file(path):

		if os.path.exists(path):
			return static_file(path)
		if os.path.exists(path + ".html"):
			return static_file(path + ".html")
		elif os.path.exists(path + ".pyhp"):
			return file(path + ".pyhp")

	run(host="::",port=1337)



import datetime
import inspect
import os

from ._internal import defaultarguments, gopen, doreahconfig

_config = {}

_queue = []
_locked = False

# set configuration
# logfolder			folder to store logfiles in
# timeformat		strftime format for log files
# defaultmodule		name for the main running script
# verbosity			higher means more (less important) messages are shown on console
def config(logfolder="logs",timeformat="%Y/%m/%d %H:%M:%S",defaultmodule="main",verbosity=0):
	"""Configures default values for this module.

	These defaults define behaviour of function calls when respective arguments are omitted. Any call of this function will overload the configuration in the .doreah file of the project. This function must be called with all configurations, as any omitted argument will reset to default, even if it has been changed with a previous function call."""
	global _config
	_config["logfolder"] = logfolder
	_config["timeformat"] = timeformat
	_config["defaultmodule"] = defaultmodule
	_config["verbosity"] = verbosity


# initial config on import, set everything to default
config()






def log(*entries,module=None,heading=None,indent=0,importance=0):
	"""Logs all supplied arguments, separate line for each. Only writes to logfile if importance value is lower than the set verbosity value.

	:param string entries: All log entries, one line per entry
	:param string module: Custom category. Log entry will be prepended by this string in the console and be written to this file on disk. Defaults to actual name of the python module.
	:param integer heading: Heading category. Headings will be visually distinguished from normal entries.
	:param integer indent: Indent to be added to entry
	:param integer importance: Low means important. If this value is higher than the verbosity, entry will not be shown on console."""

	now = datetime.datetime.utcnow().strftime(_config["timeformat"])

	# log() can be used to add empty line
	if len(entries) == 0: entries = ("",)

	# make it easier to log data structures and such
	entries = tuple([str(msg) for msg in entries])

	# header formating
	if heading == 2:
		entries = ("","","####") + entries + ("####","")
	elif heading == 1:
		entries = ("","","","# # # # #","") + entries + ("","# # # # #","","")

	# indent
	prefix = "\t" * indent

	# module name
	if module is None:
		try:
			module = inspect.getmodule(inspect.stack()[1][0]).__name__
			if module == "__main__": module = _config["defaultmodule"]
		except:
			module = "interpreter"

	global _locked, _queue
	if _locked:
		for msg in entries:
			_queue.append({"time":now,"prefix":prefix,"msg":msg,"module":module,"console":(importance <= _config["verbosity"])})
	else:
		# console output
		if (importance <= _config["verbosity"]):
			for msg in entries:
				print("[" + module + "] " + prefix + msg)

		# file output
		logfilename = _config["logfolder"] + "/" + module + ".log"
		#os.makedirs(os.path.dirname(logfilename), exist_ok=True)
		with gopen(logfilename,"a") as logfile:
			for msg in entries:
				logfile.write(now + "  " + prefix + msg + "\n")


def flush():
	"""Outputs and empties the complete backlog."""
	global _queue
	for entry in _queue:
		# console output
		if entry["console"]:
			print("[" + entry["module"] + "] " + entry["prefix"] + entry["msg"])

		# file output
		logfilename = _config["logfolder"] + "/" + entry["module"] + ".log"
		#os.makedirs(os.path.dirname(logfilename), exist_ok=True)
		with gopen(logfilename,"a") as logfile:
			logfile.write(entry["time"] + "  " + entry["prefix"] + entry["msg"] + "\n")

	_queue = []

# Quicker way to add header
def logh1(*args,**kwargs):
	"""Logs a top-level header. Otherwise, same arguments as :func:`log`"""
	return log(*args,**kwargs,header=1)
def logh2(*args,**kwargs):
	"""Logs a second-level header. Otherwise, same arguments as :func:`log`"""
	return log(*args,**kwargs,header=2)




# now check local configuration file
_config.update(doreahconfig("logging"))



from threading import Thread, Timer
import datetime

from ._internal import DEFAULT, defaultarguments, gopen, doreahconfig


_config = {}

_yearly_funcs = []
_monthly_funcs = []
_daily_funcs = []

# set configuration
def config(autostart=True):
	"""Configures default values for this module.

	These defaults define behaviour of function calls when respective arguments are omitted. Any call of this function will overload the configuration in the .doreah file of the project. This function must be called with all configurations, as any omitted argument will reset to default, even if it has been changed with a previous function call."""
	global _config
	_config["autostart"] = autostart

# initial config on import, set everything to default
config()


def yearly(func):
	"""Decorator for yearly functions. Depending on configuration, is equivalent to either :meth:`runyearly` or :meth:`repeatyearly`."""
	if _config["autostart"]: return runyearly(func)
	else: return repeatyearly(func)

def monthly(func):
	"""Decorator for monthly functions. Depending on configuration, is equivalent to either :meth:`runmonthly` or :meth:`repeatmonthly`."""
	if _config["autostart"]: return runmonthly(func)
	else: return repeatmonthly(func)

def daily(func):
	"""Decorator for daily functions. Depending on configuration, is equivalent to either :meth:`rundaily` or :meth:`repeatdaily`."""
	if _config["autostart"]: return rundaily(func)
	else: return repeatdaily(func)

def hourly(func):
	"""Decorator for hourly functions. Depending on configuration, is equivalent to either :meth:`runhourly` or :meth:`repeathourly`."""
	if _config["autostart"]: return runhourly(func)
	else: return repeathourly(func)




def runyearly(func):
	"""Decorator to make the function execute on first definition as well as every year."""

	def self_scheduling_func():
		# execute function
		func()

		# schedule next execution
		now = datetime.datetime.utcnow()
		nextyear = datetime.datetime(now.year+1,1,1)
		wait = nextyear.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func).start()

	# now execute it for the first time
	t = Timer(5,self_scheduling_func)
	t.daemon = True
	t.start()
	return func


def repeatyearly(func):
	"""Decorator to make the function repeat every new year after being called once."""

	def self_scheduling_func(*args,**kwargs):
		# execute function
		func(*args,**kwargs)

		# schedule next execution
		now = datetime.datetime.utcnow()
		nextyear = datetime.datetime(now.year+1,1,1)
		wait = nextyear.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func,args=args,kwargs=kwargs).start()

	def starter(*args,**kwargs):
		t = Thread(target=self_scheduling_func,args=args,kwargs=kwargs)
		t.daemon = True
		t.start()
	return starter

def runmonthly(func):
	"""Decorator to make the function execute on first definition as well as every month."""

	def self_scheduling_func():
		# execute function
		func()

		# schedule next execution
		now = datetime.datetime.utcnow()
		nextmonth = datetime.datetime(now.year,now.month + 1,1) if now.month != 12 else datetime.datetime(now.year+1,1,1)
		wait = nextmonth.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func).start()

	# now execute it for the first time
	t = Timer(5,self_scheduling_func)
	t.daemon = True
	t.start()
	return func


def repeatmonthly(func):
	"""Decorator to make the function repeat every new month after being called once."""

	def self_scheduling_func(*args,**kwargs):
		# execute function
		func(*args,**kwargs)

		# schedule next execution
		now = datetime.datetime.utcnow()
		nextmonth = datetime.datetime(now.year,now.month + 1,1) if now.month != 12 else datetime.datetime(now.year+1,1,1)
		wait = nextmonth.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func,args=args,kwargs=kwargs).start()

	def starter(*args,**kwargs):
		t = Thread(target=self_scheduling_func,args=args,kwargs=kwargs)
		t.daemon = True
		t.start()
	return starter

def rundaily(func):
	"""Decorator to make the function execute on first definition as well as every day."""

	def self_scheduling_func():
		# execute function
		func()

		# schedule next execution
		now = datetime.datetime.utcnow()
		nextday = datetime.datetime(now.year,now.month,now.day) + datetime.timedelta(days=1)
		wait = nextday.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func).start()

	# now execute it for the first time
	t = Timer(5,self_scheduling_func)
	t.daemon = True
	t.start()
	return func


def repeatdaily(func):
	"""Decorator to make the function repeat every new day after being called once."""

	def self_scheduling_func(*args,**kwargs):
		# execute function
		func(*args,**kwargs)

		# schedule next execution
		now = datetime.datetime.utcnow()
		nextday = datetime.datetime(now.year,now.month,now.day) + datetime.timedelta(days=1)
		wait = nextday.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func,args=args,kwargs=kwargs).start()

	def starter(*args,**kwargs):
		t = Thread(target=self_scheduling_func,args=args,kwargs=kwargs)
		t.daemon = True
		t.start()
	return starter



def runhourly(func):
	"""Decorator to make the function execute on first definition as well as every hour."""

	def self_scheduling_func():
		# execute function
		func()

		# schedule next execution
		now = datetime.datetime.utcnow()
		nexthour = datetime.datetime(now.year,now.month,now.day,now.hour) + datetime.timedelta(hours=1)
		wait = nexthour.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func).start()

	# now execute it for the first time
	t = Timer(5,self_scheduling_func)
	t.daemon = True
	t.start()
	return func


def repeathourly(func):
	"""Decorator to make the function repeat every new hour after being called once."""

	def self_scheduling_func(*args,**kwargs):
		# execute function
		func(*args,**kwargs)

		# schedule next execution
		now = datetime.datetime.utcnow()
		nexthour = datetime.datetime(now.year,now.month,now.day,now.hour) + datetime.timedelta(hours=1)
		wait = nexthour.timestamp() - now.timestamp() + 5
		Timer(wait,self_scheduling_func,args=args,kwargs=kwargs).start()

	def starter(*args,**kwargs):
		t = Thread(target=self_scheduling_func,args=args,kwargs=kwargs)
		t.daemon = True
		t.start()
	return starter





#for testing
def _runoften(func):
	def self_scheduling_func():
		# execute function
		func()

		# schedule next execution
		wait = 15
		Timer(wait,self_scheduling_func).start()


	# now execute it for the first time

	t = Timer(5,self_scheduling_func)
	t.daemon = True
	t.start()
	return func


def _repeatoften(func):
	def self_scheduling_func(*args,**kwargs):
		# execute function
		func(*args,**kwargs)

		# schedule next execution
		wait = 15
		Timer(wait,self_scheduling_func,args=args,kwargs=kwargs).start()

	def starter(*args,**kwargs):
		t = Thread(target=self_scheduling_func,args=args,kwargs=kwargs)
		t.daemon = True
		t.start()
	return starter


# now check local configuration file
_config.update(doreahconfig("regular"))



import math
import datetime
from threading import Thread
import os
import sys
import pickle

from ._internal import DEFAULT, defaultarguments, doreahconfig
from .persistence import save, load, delete, size
from .regular import repeathourly

_config = {}

_caches = {}

def config(maxsize=math.inf,maxmemory=math.inf,maxstorage=16*1024*1024*1024,maxage=60*60*24*1,maxage_negative=60*60*24*1,lazy_refresh=True,folder="cache"):
	"""Configures default values for this module.

	These defaults define behaviour of function calls when respective arguments are omitted. Any call of this function will overload the configuration in the .doreah file of the project. This function must be called with all configurations, as any omitted argument will reset to default, even if it has been changed with a previous function call."""
	global _config
	_config["maxsize"] = maxsize
	_config["maxmemory"] = maxmemory
	_config["maxstorage"] = maxstorage
	_config["maxage"] = maxage
	_config["maxage_negative"] = maxage_negative
	_config["folder"] = folder
	if _config["maxsize"] is None: _config["maxsize"] = math.inf
	if _config["maxage"] is None: _config["maxage"] = math.inf
	if _config["maxage_negative"] is None: _config["maxage_negative"] = math.inf
	_config["lazy_refresh"] = lazy_refresh


config()


class Cache:
	"""Dictionary-like object to store key-value pairs up to a certain amount and discard them after they expire.

	:param integer maxsize: Amount of entries the cache should hold before discarding old entries
	:param integer maxmemory: Soft memory limit for the cache (in bytes), not meant for precision
	:param integer maxage: Time in seconds entries are valid for after their last update. Entries older than this value are lazily removed, which means they might still be accessible with the ``allow_expired`` argument of the :meth:`get` method.
	:param integer maxage_negative: Time in seconds entries with the ``None`` value are valid. This is useful for negative caching.
	:param boolean persistent: Whether this cache should be persistent through program restarts
	:param string name: If persistent, this is the filename used"""

	@defaultarguments(_config,maxsize="maxsize",maxmemory="maxmemory",maxage="maxage",maxage_negative="maxage_negative")
	def __init__(self,maxsize=DEFAULT,maxmemory=DEFAULT,maxage=DEFAULT,maxage_negative=DEFAULT,persistent=False,name=None):

		self.maxsize = maxsize
		self.maxmemory = maxmemory
		self.maxage = maxage
		self.maxage_negative = maxage_negative
		self.persistent = persistent
		if self.persistent:
			self.name = name
			try:
				obj = load(name,folder=_config["folder"])
				if obj is not None:
					self.cache,self.times = obj
					self.changed = False
					self._maintenance()
					return
			except:
				pass
		# if either no object loaded, or not persistent in the first place
		self.cache = {}
		self.times = {}
		self.changed = False
		self._maintenance()

	def __del__(self):
		self._maintenance()

#	def create(name="defaultcache",**kwargs):
#		"""Create a new Cache object, preinitializing it from disk if applicable
#
#		:param string name: Name of the object, consistent between sessions
#		:param kwargs: Standard arguments of :class:`Cache` object creation
#		:return: Newly created object"""
#
#		obj = load(name,folder=_config["folder"])
#		if obj is not None: return obj
#
#		obj = Cache(**kwargs,persistent=True,name=name)
#		return obj

	def __contains__(self,key):
		#if key not in self.cache: return False

		#now_stamp = int(datetime.datetime.utcnow().timestamp())
		#cache_stamp = self.times.get(key)

		#if self.cache[key] is None and (now_stamp - cache_stamp) < self.maxage_negative: return True
		#if self.cache[key] is not None and (now_stamp - cache_stamp) < self.maxage: return True

		#return False
		try:
			self.get(key)
			return True
		except KeyError:
			return False


	def get(self,key,allow_expired=False):
		"""Get the value of a key in the cache.

		:param key: Key to be retrieved
		:param boolean allow_expired: If set to True, entries that have already expired will still be returned.
		:return: Value of the requested key in the cache
		:raises KeyError: No valid entry for the key found."""

		if key not in self.cache: raise KeyError()

		value = self.cache[key]

		if allow_expired: return value

		now_stamp = int(datetime.datetime.utcnow().timestamp())
		cache_stamp = self.times.get(key)



		if value is None and (now_stamp - cache_stamp) < self.maxage_negative: return value
		if value is not None and (now_stamp - cache_stamp) < self.maxage: return value

		raise KeyError()

	def add(self,key,value):
		"""Add an entry to the cache.

		:param key: Key to be added
		:param value: Value of this entry"""

#		if len(self.cache) < self.maxsize: #still have space
#			pass
#		else:
#			self.flush() #try to make space
#			if len(self.cache) < self.maxsize: pass # success
#			else:
#				self.
#				raise MemoryError() # still no space


		self.cache[key] = value
		self.times[key] = int(datetime.datetime.utcnow().timestamp())

		self._onupdate()

	def flush(self):
		"""Flush all expired entries from the cache. This is normally done lazily when needed, so this function does not need to be called manually."""

		now_stamp = int(datetime.datetime.utcnow().timestamp())

		for key in list(self.cache.keys()):
			value = self.cache[key]
			cache_stamp = self.times.get(key)

			if value is None and (now_stamp - cache_stamp) > self.maxage_negative:
				del self.cache[key]
				del self.times[key]
			if value is not None and (now_stamp - cache_stamp) > self.maxage:
				del self.cache[key]
				del self.times[key]


	def _onupdate(self):
		self.changed = True

	@repeathourly
	def _maintenance(self):
		if self.changed:
			if len(self.cache) > self.maxsize or self._size() > self.maxmemory:
				#flush anyway expired entries
				self.flush()

			while len(self.cache) > self.maxsize or self._size() > self.maxmemory:
				#expire oldest entry
				keys = list(self.times.keys())
				keys.sort(key=lambda k:self.times[k])
				delkey = keys[0]
				del self.cache[delkey]
				del self.times[delkey]

			if self.persistent:
				save((self.cache,self.times),self.name,folder=_config["folder"])

			self.changed = False



	def _size(self):
		return sys.getsizeof(pickle.dumps(self.cache))



class _DiskReference:
	def __init__(self,filename):
		self.filename = filename

class DeepCache(Cache):
	"""Dictionary-like object to store key-value pairs with the help of hard disk storage up to a certain amount and discard them after they expire.

	:param integer maxmemory: Soft memory limit for the cache (in bytes), not meant for precision
	:param integer maxstorage: Soft disk space limit for the cache (in bytes), not meant for precision
	:param integer maxage: Time in seconds entries are valid for after their last update. Entries older than this value are lazily removed, which means they might still be accessible with the ``allow_expired`` argument of the :meth:`get` method.
	:param integer maxage_negative: Time in seconds entries with the ``None`` value are valid. This is useful for negative caching.
	:param string name: Directory name used for storage"""

	@defaultarguments(_config,maxmemory="maxmemory",maxstorage="maxstorage",maxage="maxage",maxage_negative="maxage_negative")
	def __init__(self,maxmemory=DEFAULT,maxage=DEFAULT,maxstorage=DEFAULT,maxage_negative=DEFAULT,name="deepcache"):

		self.maxmemory = maxmemory
		self.maxstorage = maxstorage
		self.maxage = maxage
		self.maxage_negative = maxage_negative

		self.name = name

		try:
			obj = load(self._file(),folder=_config["folder"])
			self.cache,self.times,self.counter = obj
			self.changed = False
		except:
			self.cache = {}
			self.times = {}
			self.counter = 0

		self.changed = False

		self._maintenance()

	def get(self,key,allow_expired=False):
		"""Get the value of a key in the cache.

		:param key: Key to be retrieved
		:param boolean allow_expired: If set to True, entries that have already expired will still be returned.
		:return: Value of the requested key in the cache
		:raises KeyError: No valid entry for the key found."""

		if key not in self.cache: raise KeyError()

		value = self.cache[key]
		if isinstance(value,_DiskReference):
			value = load(self._file(value.filename),folder=_config["folder"])
		if allow_expired: return value

		now_stamp = int(datetime.datetime.utcnow().timestamp())
		cache_stamp = self.times.get(key)



		if value is None and (now_stamp - cache_stamp) < self.maxage_negative: return value
		if value is not None and (now_stamp - cache_stamp) < self.maxage: return value

		raise KeyError()

	def add(self,key,value):
		"""Add an entry to the cache.

		:param key: Key to be added
		:param value: Value of this entry"""

		# remove old file
		if key in self.cache and isinstance(self.cache[key],_DiskReference):
			delete(self._file(self.cache[key].filename),folder=_config["folder"])

		self.cache[key] = value
		self.times[key] = int(datetime.datetime.utcnow().timestamp())

		self._onupdate()

	def flush(self):
		"""Flush all expired entries from the cache. This is normally done lazily when needed, so this function does not need to be called manually."""

		now_stamp = int(datetime.datetime.utcnow().timestamp())

		for key in list(self.cache.keys()):
			value = self.cache[key]
			cache_stamp = self.times.get(key)

			if value is None and (now_stamp - cache_stamp) > self.maxage_negative:
				if isinstance(self.cache[key],_DiskReference): delete(self._file(self.cache[key].filename),folder=_config["folder"])
				del self.cache[key]
				del self.times[key]
			if value is not None and (now_stamp - cache_stamp) > self.maxage:
				if isinstance(self.cache[key],_DiskReference): delete(self._file(self.cache[key].filename),folder=_config["folder"])
				del self.cache[key]
				del self.times[key]


	def _onupdate(self):
		self.changed = True

	@repeathourly
	def _maintenance(self):
		if self.changed:
			if self._size() > self.maxmemory:
				#flush anyway expired entries
				self.flush()

			while self._size() > self.maxmemory:
				#serialize biggest entry
				keys = list(self.times.keys())
				keys.sort(key=lambda k:sys.getsizeof(pickle.dumps(self.cache[k])),reverse=True)
				keys = [k for k in keys if not isinstance(self.cache[k],_DiskReference)]
				if sys.getsizeof(pickle.dumps(self.cache[keys[0]])) < (512 * 1024):
					break
					# don't serialize tiny stuff
				try:
					movekey = keys[0]
					self._memorytodisk(movekey)
				except:
					break

			while self._disksize() > self.maxstorage:
				print("Disk size " + str(self._disksize()))
				keys = list(self.times.keys())
				keys.sort(key=lambda k:self.times[k])
				keys = [k for k in keys if isinstance(self.cache[k],_DiskReference)]
				try:
					delkey = keys[0]
					print("Deleting " + delkey)
					delete(self._file(self.cache[delkey].filename),folder=_config["folder"])
					del self.cache[delkey]
					del self.times[delkey]
				except:
					break


			save((self.cache,self.times,self.counter),self._file(),folder=_config["folder"])

			self.changed = False

	def _memorytodisk(self,key):
		print("Moving " + key + " from memory to disk")
		self.counter += 1
		value = self.cache[key]
		filename = str(self.counter)
		save(value,self._file(filename),folder=_config["folder"])
		self.cache[key] = _DiskReference(filename)

	def _disktomemory(self,key):
		print("Moving " + key + " from disk to memory")
		filename = self.cache[key].filename
		value = load("./" + self.name + "/" + filename,folder=_config["folder"])
		self.cache[key] = value
		delete(self._file(filename),folder=_config["folder"])


	def _size(self):
		return sys.getsizeof(pickle.dumps([self.cache[key] for key in self.cache if not isinstance(self.cache[key],_DiskReference)]))

	def _disksize(self):
		#return sum(os.path.getsize("./" + _config["folder"] + "/" + self.name + "/" + f) for f in os.listdir("./" + _config["folder"] + "/" + self.name))
		sumsize = 0
		for key in self.cache:
			if isinstance(self.cache[key],_DiskReference):
				sumsize += size(self._file(self.cache[key].filename),folder=_config["folder"])
		return sumsize

	# gets a relative filename for keys of this deepcache object
	def _file(self,name=None):
		if name is None: name = "root"
		return os.path.join(self.name,name)


# decorator
@defaultarguments(_config,maxsize="maxsize",maxage="maxage",maxage_negative="maxage_negative",lazy_refresh="lazy_refresh")
def cached(maxsize=DEFAULT,maxage=DEFAULT,maxage_negative=DEFAULT,lazy_refresh=DEFAULT):
	"""Method decorator to add a proxy cache to a function without keyword arguments.

	:param integer maxsize: Amount of entries the cache should hold before discarding old entries.
	:param integer maxage: Time in seconds entries are valid for after their last update. Entries older than this value are lazily removed, which means they might still be accessible with the ``allow_expired`` argument of the :meth:`get` method.
	:param integer maxage_negative: Time in seconds entries with the ``None`` value are valid. This is useful for negative caching.
	:param boolean lazy_refresh: If True, expired cache entries will still be returned, but trigger a background refresh. Useful if speed is more important than currentness of data."""


	# decorator for lazy refresh
	if lazy_refresh:
		def cacheddecorator(func):

			def newfunc(*args,cache=Cache(maxsize=maxsize,maxage=maxage,maxage_negative=maxage_negative)):
				try:
					result = cache.get(args,allow_expired=True)
				except:
					result = func(*args)
					cache.add(args,result)
					return result

				if args not in cache: # if we just got an expired result, we want to execute the function and add to cache

					def fillcache(cache,args,func):
						cache.add(args,func(*args))

					t = Thread(target=fillcache,args=(cache,args,func,))
					t.start()

				return result

			return newfunc

	# decorator for non-lazy
	else:
		def cacheddecorator(func):

			def newfunc(*args,cache=Cache(maxsize=maxsize,maxage=maxage,maxage_negative=maxage_negative)):
				try:
					return cache.get(args)
				except:
					result = func(*args)
					cache.add(args,result)
					return result

			return newfunc

	return cacheddecorator



# now check local configuration file
_config.update(doreahconfig("caching"))
